import logging
import requests
from urllib.parse import urljoin
from requests_kerberos import HTTPKerberosAuth
from datetime import datetime
from pytz import timezone

LOGGER = logging.getLogger('hadoop_nm_fix')

class NodeManagerScanner(object):

    def __init__(self, rm_fqdn: str, threshold_mins: int):
        self.rm_fqdn = rm_fqdn
        self.threshold_mins = threshold_mins

    def get_node_managers(self, states):

        response = requests.get(urljoin('https://' + self.rm_fqdn + ':8090', 'ws/v1/cluster/nodes'),
                                params={'states': states},
                                auth = HTTPKerberosAuth())
        if response.status_code == requests.codes.ok:
            return response.json()['nodes']['node']
        else:
            response.raise_for_status()

    def get_containers(self, node_fqdn):

        response = requests.get(urljoin('https://' + node_fqdn + ':8042', 'ws/v1/node/containers'),
                                auth = HTTPKerberosAuth())
        if response.status_code == requests.codes.ok:
            if response.json()['containers']:
                return response.json()['containers']['container']
            else:
                return []
        else:
            response.raise_for_status()

    def is_node_healthy(self, node_fqdn):
        response = requests.get(urljoin('https://' + node_fqdn + ':8042', 'ws/v1/node'),
                                auth = HTTPKerberosAuth())
        if response.status_code == requests.codes.ok:
            if response.json()['nodeInfo']['nodeHealthy'] == True:
                LOGGER.debug(
                    f'Nodemanager {node_fqdn} is healthy')
                return True
            else: 
                LOGGER.debug(
                    f'Nodemanager {node_fqdn} is NOT healthy')
                return False
        else:
            response.raise_for_status()

    def is_node_manager_stalled(self, container):
        container_id = container['id']
        container_state = container['state']
        if container_state == 'KILLING':
            container_diagnotics = container['diagnostics']
            timestamp = NodeManagerScanner.parse_diagnostics(container['diagnostics'])
            if timestamp is not None:
                local_time_now = datetime.now(timezone('Europe/Brussels')).replace(tzinfo=None)
                timestamp_diff_mins = (local_time_now - timestamp).seconds / 60
                LOGGER.debug(
                    f'Container {container_id} is in state {container_state} for {timestamp_diff_mins} minutes')
                return timestamp_diff_mins > self.threshold_mins
            else:
                LOGGER.debug(
                    f'Container {container_id} is in state {container_state}, but probably not killed by RM: {container_diagnotics}')
        else:
            LOGGER.debug(f'Container {container_id} is in state {container_state}')
            return False


    @staticmethod
    def parse_diagnostics(diagnostics: str):
        for diagnostic in diagnostics.split('\n'):
            if 'container killed by resourcemanager' in diagnostic.lower():
                timestamp_str = diagnostic[1:20] # 2022-07-18 10:44:32
                # sadly, timestamps returned by YARN REST API's are not timezone aware
                return datetime.strptime(timestamp_str, '%Y-%m-%d %H:%M:%S')
        else:
            return None


if __name__ == '__main__':

    import os
    import json
    import logging
    import logging.config
    import argparse

    _basedir = os.path.dirname(os.path.realpath(__file__))
    logging.config.fileConfig(os.path.join(_basedir, 'logging.conf'),
                              disable_existing_loggers=False)

    parser = argparse.ArgumentParser(description='Scan stalled Hadoop NodeManagers.')
    parser.add_argument('rm_fqdn', type=str,
                        help='ResourceManager FQDN')
    parser.add_argument('threshold_mins', type=int,
                        help='Threshold for detecting killing containers, in minutes')

    args = parser.parse_args()

    nm_scanner = NodeManagerScanner(args.rm_fqdn,
                                    args.threshold_mins)

    stalled_nms = []

    nms = nm_scanner.get_node_managers('RUNNING')
    for nm in nms:
        nm_fqdn = nm['nodeHostName']
        LOGGER.debug(f'Scanning NodeManager {nm_fqdn}')
        if nm_scanner.is_node_healthy(nm_fqdn):
          containers = nm_scanner.get_containers(nm_fqdn)
          for container in containers:
              if nm_scanner.is_node_manager_stalled(container):
                  stalled_nms.append(nm_fqdn)
                  break

    print(json.dumps(stalled_nms))
